/*
compilation:
g++ `pkg-config --libs opencv --cflags opencv` -lglut -lGLU <filename>.cpp -o <output file>
*/

/*OpenGL - include headers in this order!*/
#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>

/*OpenCV*/
#include "cv.h"
#include "highgui.h"


/*global variables*/
float widthscale = 0.05, heightscale = 0.05;	//how many OpenGL coordinate units per pixel?
int win_width = 1000, win_height = 1000;			//initial window size in pixels
float move_amnt = 0.5;							//hero move amount in OpenGL coordinates

int loadTexture_Ipl(IplImage *image, GLuint *text) {
/*modified code from http://blog.damiles.com/?p=9*/

	if (image==NULL) return -1;

	glGenTextures(1, text);

	glBindTexture( GL_TEXTURE_2D, *text ); //bind the texture to its array
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image->width, image->height,0, GL_BGR_EXT, GL_UNSIGNED_BYTE, image->imageData);

	return 0;
}


class visual{
/*The visual class is a base class for all other classes that use images, such as background objects
and sprites. It contains methods for specifying the image, the location (center) of the image in OpenGL
coordinates, orientation and scale, and drawing the image.*/

	public:
		GLuint tex; 			//texture
		int pixwidth, pixheight;
		float width, height;	//image width and height in OpenGL coordinates
		float xscale, yscale;	//scaling factors
		float x, y, z, theta;	//OpenGL coords of image center + rotation
		int visible;

	visual() {
		xscale = 1;
		yscale = 1;
		visible = 1;
	}

	int setImage(const char string[]){
	/*set the image*/
		IplImage *image = NULL;
		image = cvLoadImage(string);
		if(!image)
			return 1;	//something went wrong
		pixwidth = image->width;
		pixheight = image->height;
		width = pixwidth*widthscale;		//in OpenGL coordinates
		height = pixheight*heightscale;	//in OpenGL coordinates
		loadTexture_Ipl(image, &tex);
		cvReleaseImage(&image);
		return 0;
	}

	int setXYZ(float xx, float yy, float zz){
	/*set the x,y,z coords of image center*/
		x = xx; y = yy; z = zz;
		return 0;
	}

	int setTheta(float t){
	/*set image rotation - degrees counterclockwise*/
		theta = t;
		return 0;
	}


	int setVisible(int v){
	/*set visibility status of object*/
		visible = v;
		return 0;
	}

	int setScale(float xs, float ys){
	/*set scaling coefficients*/
		xscale = xs;
		yscale = ys;
		return 0;
	}

	virtual int draw(){
	/*perform OpenGL calls to draw image - virtual to allow masking operations in sprite class*/
		float ww, hh;

		ww = width*xscale*0.5;
		hh = height*yscale*0.5;

		if(!visible){	//we're coastin' for a ghostin'
			return 0;	//so don't do anything
		}

		glPushMatrix();
		glTranslatef(x,y,0);
		glRotatef(theta,0,0,1);
		glBindTexture(GL_TEXTURE_2D, tex);
	    glBegin(GL_QUADS);
			glTexCoord2f(0.0, 1.0); glVertex3f(-ww, -hh, z);	//bottom left
			glTexCoord2f(1.0, 1.0); glVertex3f(ww, -hh, z);		//bottom right
			glTexCoord2f(1.0, 0.0); glVertex3f(ww, hh, z);		//top right
			glTexCoord2f(0.0, 0.0); glVertex3f(-ww, hh, z);		//top left
		glEnd();
		glPopMatrix();
		return 0;
	}

};

class background: public visual{
/*visual with a second b/w image that is used to determine boundaries through which sprites don't pass*/
	public:
		IplImage *bndimage;

	int setBoundaryImage(char* bndstring){
		bndimage = cvLoadImage(bndstring);	
		if(!bndimage)
			return 1;	//error
		return 0;
	}
};


class maskvisual: public visual{
/*modified visual class that works with masks to create "odd" shaped visuals w/ help of blending*/ 
	public:
		GLuint masktex;

	int setMask(const char string[]){
		/*set mask image - basically same code as for setImage method in visual class*/
		IplImage *image = NULL;
		image = cvLoadImage(string);
		if(!image)
			return 1;	//something went wrong
		loadTexture_Ipl(image, &masktex);
		cvReleaseImage(&image);
		return 0;	
	}

	int draw(){
	/*differs slightly from virtual method in visual class*/

		float ww, hh;

		ww = width*xscale*0.5;
		hh = height*yscale*0.5;

		if(!visible){
			return 0;
		}

		glPushMatrix();
		glTranslatef(x,y,0);
		glRotatef(theta,0,0,1);

		glEnable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);

		glBlendFunc(GL_ONE_MINUS_SRC_COLOR,GL_SRC_COLOR);

			glBindTexture(GL_TEXTURE_2D, masktex);
	    	glBegin(GL_QUADS);
			glTexCoord2f(0.0, 1.0); glVertex3f(-ww, -hh, z);	//bottom left
			glTexCoord2f(1.0, 1.0); glVertex3f(ww, -hh, z);		//bottom right
			glTexCoord2f(1.0, 0.0); glVertex3f(ww, hh, z);		//top right
			glTexCoord2f(0.0, 0.0); glVertex3f(-ww, hh, z);		//top left
		glEnd();

		glBlendFunc(GL_SRC_COLOR,GL_ONE);

		glBindTexture(GL_TEXTURE_2D, tex);
    	glBegin(GL_QUADS);
			glTexCoord2f(0.0, 1.0); glVertex3f(-ww, -hh, z);	//bottom left
			glTexCoord2f(1.0, 1.0); glVertex3f(ww, -hh, z);		//bottom right
			glTexCoord2f(1.0, 0.0); glVertex3f(ww, hh, z);		//top right
			glTexCoord2f(0.0, 0.0); glVertex3f(-ww, hh, z);		//top left
		glEnd();

		glPopMatrix();

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		return 0;
	}
};

class sprite: public maskvisual{
/*extension of maskvisual class that adds hitbox functionality and virtual onHit method.
Possible future extension: record data about what hit the sprite, not just that it was hit -
could do this by assigning each sprite or sprite-daughter class a unique int classID attribute
and use the classID in a switch statement in the onHit method - might not need this functionality,
though*/
	public:
		/*hit box corner positions relative to sprite's center
		naming example: hbt = "HitBox Top (distance from center)" */
		float hbt, hbb, hbl, hbr;
		int facing;	//0 up, 1 right, 2 down, 3 left
		sprite* othersprite;	//pointer to the last sprite that this sprite was hit by (or NULL)

	int setHitBox(float t, float b, float l, float r){
	/*have to set boundaries before any calls to isHitBy otherwise it is impossible to detect hit*/
		hbt = t; hbb = b; hbl = l; hbr = r;	//hbt = "hit box top" boundary
		return 0;
	}

	int isHitBy(sprite *sp){
	/*only interactive things (sprites or any daughter classes) should be able to hit a sprite
	(or any daughter class). sp is the sprite we're checking to see if we're hit by.
	Ex. character.isHitBy(enemy) would check to see if sprite character is hit by sprite enemy.*/ 
		float spx = sp->x, spy = sp->y, spt = sp->hbt, spb = sp->hbb, spl = sp->hbl, spr = sp->hbr,
				spxscale = sp->xscale, spyscale = sp->yscale;

		if(spx+spr*spxscale > x-hbl*xscale && spx-spl*spxscale < x+hbr*xscale && 
			spy+spt*spyscale > y-hbb*yscale && spy-spb*spyscale < y+hbt*yscale){
			othersprite = sp;
			return 1;
		}
		else{
			othersprite = NULL;
			return 0;
		}
	}

	virtual int onHit(){
		/*print out hit information as default behavior. this will be overridden by the onHit() method
		in actual derived classes*/
		printf("sprite %lu was hit\n", (long unsigned int)this);
		return 0;
	}

	int setFacing(int f){
	/*self explanatory, really...*/
		facing = f;
		return 0;
	}

	int trySetXYZ(float xx, float yy, float zz, background* bg){
	/*try to set xyz. fails if there is a background obstacle in the way*/
		int pixelx, pixely;
		int bndval;
		pixelx = xx/widthscale;
		pixelx += bg->pixwidth*0.5*bg->xscale;
		pixely = bg->pixheight*0.5*bg->yscale;
		pixely -= yy/heightscale;

		pixelx /= bg->xscale;
		pixely /= bg->yscale;

		//check obstacle
		//printf("\npixel (x,y) = (%d,%d)\n",pixelx,pixely);
		//printf("\nval = %d\n", (uchar)*(bg->bndimage->imageData + pixely*bg->bndimage->widthStep +pixelx*bg->bndimage->nChannels));
		bndval = (uchar)*(bg->bndimage->imageData + pixely*bg->bndimage->widthStep +pixelx*bg->bndimage->nChannels);
		if(bndval){
			x = xx;


			y=yy;
			return 0;	//set succeeded
		}

		return 0;	//set failed
	}
};


class obstacle: public sprite{
/*interacts with sprites, and so must also be a sprite according to my arbitrary rules. OBEY OBEY OBEY.
Really, what separates this class from sprite is a specific onHit() function.*/
	public:
	
	int onHit(){
	/*block any sprite that hits it, so the onHit action is registered in obstacle, not whatever hits obstacle (saves code).*/	
		
		switch(othersprite->facing){
			case 0:{
				othersprite->y -= move_amnt;
				break;
			}

			case 1:{
				othersprite->x -= move_amnt;
				break;
			}

			case 2:{
				othersprite->y += move_amnt;
				break;
			}

			case 3:{
				othersprite->x += move_amnt;
				break;
			}

			default: break;
		}

	/*spin to show your anger at being hit!*/
		theta += 15;
	}	
};



/*we'll make these global for this test*/
sprite hero;	//sprite
obstacle enemy;	//test hit detection
background bg;	//background


void init(void)
{
	glClearColor(0.2,0.4,0.8,0); //clear color 
	glShadeModel(GL_FLAT);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);

	hero.setMask("herman_mask.bmp");
	hero.setImage("herman.bmp");
	hero.setHitBox(10,10,10,10);
	hero.setScale(.1,.1);

	enemy.setMask("sprite_mask.bmp");
	enemy.setImage("sprite.bmp");
	enemy.setXYZ(11,-9,1);
	enemy.setHitBox(2,2,2,2);
	enemy.setScale(.5,.5);

	bg.setImage("test_im.bmp");
	bg.setBoundaryImage("test_im_bnd.bmp");
	bg.setScale(1.5,1.5);

	return;
}



void display(void)
{

	/*define view*/
	glLoadIdentity();
	gluLookAt(0.0,0.0,50.0,0.0,0.0,0.0,0.0,1.0,0.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/*check for hits BEFORE drawing (avoids jittery movement if keyboard interrupt executes before
	isHitBy checks are run)*/
	if(enemy.isHitBy(&hero)){
		enemy.onHit();
	}
	if(hero.isHitBy(&enemy)){
		hero.onHit();
	}

	bg.draw();
	enemy.draw();
	hero.draw();

	glutPostRedisplay(); //force redraw
	glutSwapBuffers();
}



void keyboard(unsigned char key, int x, int y)
{
	switch(key){
		case 'w':{
			hero.trySetXYZ(hero.x,hero.y+move_amnt,hero.z,&bg);
			if(hero.facing){
				hero.setFacing(0);
				hero.setTheta(0);
			}
			break;
		}
		case 'a':{
			hero.trySetXYZ(hero.x-move_amnt,hero.y,hero.z,&bg);
			if(hero.facing != 3){
				hero.setFacing(3);
				hero.setTheta(90);
			}
			break;
		}
		case 's':{
			hero.trySetXYZ(hero.x,hero.y-move_amnt,hero.z,&bg);
			if(hero.facing != 2){
				hero.setFacing(2);
				hero.setTheta(180);
			}
			break;
		}
		case 'd':{
			hero.trySetXYZ(hero.x+move_amnt,hero.y,hero.z,&bg);
			if(hero.facing != 1){
				hero.setFacing(1);
				hero.setTheta(-90);
			}
			break;
		}
		default:
		    break;
	}
}



void reshape(int w, int h)
{
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-w*widthscale*0.5,w*widthscale*0.5,-h*widthscale*0.5,h*widthscale*0.5,0.0,60.0);
	glMatrixMode(GL_MODELVIEW);
}



int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize (win_width, win_height);
	glutInitWindowPosition (100, 100);
	glutCreateWindow (argv[0]);
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);

	glutMainLoop();

	return 0;
}

